#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <ctype.h>

#include "btserverdefines.h"
#include "bthandler.h"
#include "btutils.h"

#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>

//PLEASECHECK (Victor): quick hack to skip including gotype.h
#define GOTYPE_VALENCIA500	8
#define GOTYPE_VALENCIA700	9
#define GOTYPE_MURCIA500	12
#define GOTYPE_MURCIA700	13
#define GOTYPE_MILAN		18
#define GOTYPE_CAGLIARI 24

#ifdef ENABLE_DEBUG_LOG

#define STATE2STR(state) state2str(state)
#define ERROR2STR(error) error2str(error)

#define UPDATE_STATE(state, newstate) {DEBUG_LOG2("state %s -> %s", STATE2STR(state), STATE2STR(newstate)); state = newstate;}
/* Debug buffer */
#define BT_DBGBUF_LEN 4096
static char dbgbuf[BT_DBGBUF_LEN];

#else //ENABLE_DEBUG_LOG

#define STATE2STR(state) "DEBUG DISABLED"
#define ERROR2STR(error) "DEBUG DISABLED"

#define UPDATE_STATE(state, newstate) {state = newstate;}
#endif //ENABLE_DEBUG_LOG

static void handle_connection_lost(BTConnection con);

#define INIT_LISTEN_TIMER(c, t)  {BTSetTimer(&c->listentotaltr, t);BTSetTimer(&c->listenretrytr, 0);}
#define INIT_CONNECT_TIMER(c, t) {BTSetTimer(&c->connecttotaltr, t);BTSetTimer(&c->connectretrytr, 0);}


int BTHandlerCompare(cBTConnection con, const char *bdaddr, int chan, int profile, int *res)
{
  bdaddr_t addr;

  if (!con || !bdaddr || !res)
    return BT_ERR_ARGUMENT;

  str2ba(bdaddr, &addr);
  // compare address and profile
  if (!bacmp(&addr, &con->sock->bdaddr) && con->profile == profile)
    *res =  BT_TRUE;
  else
    *res = BT_FALSE;

  DEBUG_LOG4("%s, %d, %d res=%d", bdaddr, chan, profile, *res);
  (void) chan; /* silence unused variable warning */
  return BT_ERR_NONE;
}

int BTHandlerCreate(BTConnection *con, const char *bdaddr, int chan, int profile, int flags, const char *pin)
{
  DEBUG_LOG5("addr=[%s], chan=%d, prof=%d, flags=%d pin=[%s]", bdaddr, chan, profile, flags, pin);

  BTConnection c;

#ifdef FORCE_CH_SEARCH
  chan = -1;
  DEBUG_LOG(" Force channel search enabled");
#endif

  if (!con || chan > 30 || !bdaddr || !pin)
    return BT_ERR_ARGUMENT;
  if (strlen(bdaddr) > (BT_BDADDR_LEN - 1) || strlen(pin) > (BT_PIN_LEN - 1))
    return BT_ERR_ARGUMENT;
  c = (BTConnection)malloc(sizeof(struct sBTConnection));
  if (!c)
    return BT_ERR_NOMEMORY;
  if (BTSocketCreate(&c->sock, bdaddr, chan) != BT_ERR_NONE) {
    DEBUG_LOG("creating socket failed");
    free(c);
    return BT_ERR_FAILED;
  }

  c->state = BT_STATE_UNINITIALIZED;
  c->ppp.state = BT_PPP_STATE_UNINITIALIZED;
  c->profile = profile;
  c->flags = flags;
  strcpy(c->pin, pin);
  c->sendbuf = NULL;
  c->sendbuflen = 0;
  c->error = BT_ERR_NONE;
  if (flags & BT_FLAG_LISTEN) {
    INIT_LISTEN_TIMER(c, BT_TO_LISTEN_TOTAL);
    UPDATE_STATE(c->state, BT_STATE_START_LISTENING);
  } else if (flags & BT_FLAG_OPEN) {
    if (profile == BT_PROFILE_HS) {
      INIT_CONNECT_TIMER(c, BT_TO_CONNECT_TOTAL_HS);
    } else {
      INIT_CONNECT_TIMER(c, BT_TO_CONNECT_TOTAL);
    }
    UPDATE_STATE(c->state, BT_STATE_START_CONNECTING);
  } else
    UPDATE_STATE(c->state, BT_STATE_INITIALIZED);
  UPDATE_STATE(c->ppp.state, BT_PPP_STATE_UNINITIALIZED);
  c->ppp.refcount = 0;
  c->ppp.pid = -1;
  c->reconnecting = 0;
  c->fake = 0;
  c->chan_retrive_pid = -1;
  c->chan_retrive_pid_ppp = -1;
  c->hcitool_pid = -1;

  *con = c;
  return BT_ERR_NONE;
}

int BTHandlerCreateRequest(BTConnection *con)
{
  DEBUG_LOG("Entered.");

  BTConnection c;

  c = (BTConnection)malloc(sizeof(struct sBTConnection));
  if (!c)
    return BT_ERR_NOMEMORY;
  c->state = BT_STATE_START_CONNECTING;
  c->fake = 1;
  *con = c;
  return BT_ERR_NONE;
}

int BTHandlerDestroy(BTConnection con)
{
  DEBUG_LOG1("state: %s", STATE2STR(con->state));

  int res;

  if (!con)
    return BT_ERR_ARGUMENT;
  if (con->state != BT_STATE_CLOSING && con->state != BT_STATE_UNINITIALIZED)
    res = BTHandlerClose(con);
  else
    res = BT_ERR_NONE;
  if (con)
    free(con);

  return res;
}

int BTHandlerReconnect(BTConnection con)
{
  DEBUG_LOG1("state: %s", STATE2STR(con->state));
  
  if (!con)
    return(BT_ERR_ARGUMENT);
  if (con->state != BT_STATE_CONNECTED)
    return (BT_ERR_STATE);
  else
  {
    con->flags &= ~BT_FLAG_LISTEN;
    handle_connection_lost(con);
    return (BT_ERR_NONE);
  }
}

int BTHandlerStartApp(BTConnection con, const char *param)
{
  DEBUG_LOG6("profile=%d, channel %d ,param=[%s], state=%s, refcount=%d, flags=0x%x ", con->profile, con->sock->chan, param, STATE2STR(con->state), con->ppp.refcount, con->flags);

  if (!con || !param || con->fake)
    return BT_ERR_ARGUMENT;
  if (con->profile != BT_PROFILE_DUN)
    return BT_ERR_NONE;
  if (strlen(param) > (BT_PPP_PARAM_LEN - 1))
    return BT_ERR_ARGUMENT;
  if (con->state != BT_STATE_INITIALIZED && con->state != BT_STATE_APP_START && con->state != BT_STATE_APP_READY && con->state != BT_STATE_RETRIEVING_CHANNEL && con->state != BT_STATE_RETRIEVE_CHANNEL )
    return BT_ERR_STATE;
  con->ppp.refcount++;
  if(con->ppp.refcount > 1 && con->state != BT_STATE_INITIALIZED)
    return BT_ERR_NONE;
  else if (con->ppp.refcount == 1)
    strcpy(con->ppp.param, param);
  UPDATE_STATE(con->state, BT_STATE_APP_START);
  UPDATE_STATE(con->ppp.state, BT_PPP_STATE_INIT);

  return BT_ERR_NONE;
}

int BTHandlerStopApp(BTConnection con)
{
  int pid;

  DEBUG_LOG4(" profile=%d, state=%s, refcount=%d channel %d", con->profile, STATE2STR(con->state), con->ppp.refcount, con->sock->chan);

  if (!con || con->fake)
    return BT_ERR_ARGUMENT;
  if (con->profile != BT_PROFILE_DUN)
    return BT_ERR_NONE;
  if (con->state != BT_STATE_APP_START && con->state != BT_STATE_APP_READY && con->state != BT_STATE_INITIALIZED)
    return BT_ERR_STATE;
  if (con->ppp.refcount)
    con->ppp.refcount--;
  if (con->ppp.refcount || con->state == BT_STATE_INITIALIZED)
    return BT_ERR_NONE;
  if (con->state == BT_STATE_APP_START && (con->ppp.state == BT_PPP_STATE_CHECKTTY || con->ppp.state == BT_PPP_STATE_START || con->ppp.state == BT_PPP_STATE_CHECKREADY || con->ppp.state == BT_PPP_STATE_HCITOOLCHECK || con->ppp.state == BT_PPP_STATE_HCITOOLCHECKREADY ) )
    BTSocketReleaseTTY(con->sock);

  pid = BTGetFileInt(BT_PPP_PIDFILE);
  if (pid > 0) {
    DEBUG_LOG1("terminating %d", pid);
    kill(pid, SIGTERM);
  }
  BTSetTimer(&con->ppp.removetr, BT_TO_APPREMOVE);
  UPDATE_STATE(con->state, BT_STATE_APP_CLOSING);

  return BT_ERR_NONE;
}

int BTHandlerSend(BTConnection con, const char *buf, int len, int *sent)
{
#ifdef ENABLE_DEBUG_LOG
  if (len <= BT_DBGBUF_LEN) {
    int ascii=1,i;
    if (buf && len)
      strncpy(dbgbuf, buf, len);
    dbgbuf[len - 1] = '\0';
    if (len)
    {
      for (i=0; i<len && ascii; i++)
      {
        if (!isprint(buf[i]) && buf[i] != '\n' && buf[i] != '\r') ascii=0;
      }
      if (ascii)
      {
        DEBUG_LOG3("SendBuf=[%s], len=%d, state=%s", dbgbuf, len, STATE2STR(con->state));
      }
    }
  } else {
    DEBUG_LOG("debug buffer is too small!");
  }
#endif //ENABLE_DEBUG_LOG

  char *tmpbuf;

  if (!con || !buf || !sent || con->fake)
    return BT_ERR_ARGUMENT;
  if (con->state == BT_STATE_APP_START || con->state == BT_STATE_APP_READY || con->state == BT_STATE_APP_CLOSING)
    return BT_ERR_STATE;
  // Copy data to the send buffer
  if (!con->sendbuf) {
    con->sendbuf = (char*)malloc(len);
    if (!con->sendbuf)
      return BT_ERR_NOMEMORY;
  } else {
    tmpbuf = (char*)realloc(con->sendbuf, con->sendbuflen + len);
    if (!tmpbuf)
      return BT_ERR_NOMEMORY;
    con->sendbuf = tmpbuf;
  }
  memcpy(con->sendbuf + con->sendbuflen, buf, len);
  //DEBUG_LOG2("send buffer grown %d -> %d", con->sendbuflen, con->sendbuflen + len);
  con->sendbuflen += len;
  // Connect if was not connected
  if (con->state != BT_STATE_CONNECTED) {
    if (con->state == BT_STATE_START_CONNECTING || con->state == BT_STATE_CONNECTING || con->state == BT_STATE_CONNECTED || con->state == BT_STATE_RETRIEVING_CHANNEL || con->state == BT_STATE_RETRIEVE_CHANNEL)
      return BT_ERR_NONE;
    if (con->state == BT_STATE_START_LISTENING || con->state == BT_STATE_LISTENING)
      BTSocketStopListen(con->sock);
    con->reconnecting = 0;
    if (con->profile == BT_PROFILE_HS) {
      INIT_CONNECT_TIMER(con, BT_TO_CONNECT_TOTAL_HS);
    } else {
      INIT_CONNECT_TIMER(con, BT_TO_CONNECT_TOTAL);
    }
    if (con->state != BT_STATE_RETRIEVE_CHANNEL && con->state != BT_STATE_RETRIEVING_CHANNEL)
      UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
  }
  return BT_ERR_NONE;
}

int BTHandlerReceive(BTConnection con, char *buf, int len, int *received)
{
  int res;

  if (!con || !buf || !received || con->fake)
    return BT_ERR_ARGUMENT;
  if (con->state != BT_STATE_CONNECTED)
    return BT_ERR_STATE;
  res = BTSocketReceive(con->sock, buf, len, received, BT_TO_NONE);
  switch (res) {
  case BT_ERR_TIMEOUT:
    *received = 0;
    res = BT_ERR_NONE;
    break;
  case BT_ERR_NONE:
#ifdef ENABLE_DEBUG_LOG
    if (*received <= BT_DBGBUF_LEN) {
      int ascii=1,i;
      if (*received)
      {
        strncpy(dbgbuf, buf, *received);
        buf[*received] = '\0';

        for (i=0; i<*received && ascii; i++)
        {
          if (!isprint(buf[i])) ascii=0;
        }
        if (ascii)
        {
          DEBUG_LOG3("ReceivedBuf=[%s], len=%d, received=%d", buf, len, *received);
        }
      }
    } else {
      DEBUG_LOG("debug buffer is too small!");
    }
#endif //ENABLE_DEBUG_LOG
    res = BT_ERR_NONE;
    break;
  case BT_ERR_FAILED:
    DEBUG_LOG1("receive failed: %s", strerror(errno));
    handle_connection_lost(con);
    *received = 0;
    res = BT_ERR_CONNECTION;
    break;
  default:
    DEBUG_LOG("UNHANDLED ERROR!!!");
    res = BT_ERR_FAILED;
    *received = 0;
    break;
  }
  return res;
}

int BTHandlerRelease(BTConnection con)
{
  DEBUG_LOG1("state=%s", STATE2STR(con->state));

  if (!con || con->fake)
    return BT_ERR_ARGUMENT;
  if (con->state != BT_STATE_START_CONNECTING && con->state != BT_STATE_CONNECTING && con->state != BT_STATE_CONNECTED && con->state != BT_STATE_REMOVE_CONNECTION)
    return BT_ERR_STATE;
  BTSocketClose(con->sock);
  if (con->sendbuf)
    free(con->sendbuf);
  con->sendbuf = NULL;
  con->sendbuflen = 0;
  if (con->flags & BT_FLAG_LISTEN) {
    DEBUG_LOG("release to listening");
    INIT_LISTEN_TIMER(con, BT_TO_LISTEN_TOTAL);
    UPDATE_STATE(con->state, BT_STATE_START_LISTENING);
  } else {
    DEBUG_LOG("release to initialized");
    UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
  }

  return BT_ERR_NONE;
}

int BTHandlerClose(BTConnection con)
{
  int pid;

  DEBUG_LOG1("state=%s", STATE2STR(con->state));

  if (!con)
    return BT_ERR_ARGUMENT;

  if (con->fake) {
    UPDATE_STATE(con->state, BT_STATE_UNINITIALIZED);
    return BT_ERR_NONE;
  }

  if (con->state == BT_STATE_RETRIEVING_CHANNEL || con->state == BT_STATE_RETRIEVE_CHANNEL) {
    if (con->chan_retrive_pid > 0 )
      kill(con->chan_retrive_pid, SIGTERM);

  }

  if (con->state == BT_STATE_APP_START || con->state == BT_STATE_APP_READY || con->state == BT_STATE_APP_CLOSING) {
    if (con->state != BT_STATE_APP_CLOSING) {
      pid = BTGetFileInt(BT_PPP_PIDFILE);
      if (pid > 0) {
        DEBUG_LOG1("terminating %d", pid);
        kill(pid, SIGTERM);
      }
    }
    BTSetTimer(&con->ppp.removetr, BT_TO_APPREMOVE);
    UPDATE_STATE(con->state, BT_STATE_CLOSING);
    UPDATE_STATE(con->ppp.state, BT_PPP_STATE_CLOSING);
  } else if (con->state != BT_STATE_CLOSING && con->state != BT_STATE_UNINITIALIZED) {
    BTSocketClose(con->sock);
    if (con->sendbuf)
      free(con->sendbuf);
    con->sendbuf = NULL;
    con->sendbuflen = 0;
    UPDATE_STATE(con->state, BT_STATE_CLOSING);
    UPDATE_STATE(con->ppp.state, BT_PPP_STATE_UNINITIALIZED);
  }
  return BT_ERR_NONE;
}

int BTHandlerGetState(cBTConnection con, int *state)
{
  if (!con || !state)
    return BT_ERR_ARGUMENT;

  *state = con->state;

  return BT_ERR_NONE;
}

int BTHandlerGetError(cBTConnection con, int *error)
{
  if (!con || !error)
    return BT_ERR_ARGUMENT;
  DEBUG_LOG1("error=%s", ERROR2STR(con->error));
  *error = con->error;

  return BT_ERR_NONE;
}

int BTHandlerSetFlags(BTConnection con, int flags)
{
  DEBUG_LOG3("flags=0x%X -> 0x%X, state=%s", con->flags, flags, STATE2STR(con->state));

  if (!con || con->fake)
    return BT_ERR_ARGUMENT;
  if (!(con->flags & BT_FLAG_LISTEN) && (flags & BT_FLAG_LISTEN) && con->state == BT_STATE_INITIALIZED) {
    INIT_LISTEN_TIMER(con, BT_TO_LISTEN_TOTAL);
    UPDATE_STATE(con->state, BT_STATE_START_LISTENING);
  }
  else if (!(con->flags & BT_FLAG_OPEN) && (flags & BT_FLAG_OPEN) && (con->state == BT_STATE_START_LISTENING || con->state == BT_STATE_LISTENING || con->state == BT_STATE_INITIALIZED) && !(con->flags & BT_FLAG_DONOTRECONNECT)) {
    if (con->state == BT_STATE_START_LISTENING || con->state == BT_STATE_LISTENING)
      BTSocketStopListen(con->sock);
    con->reconnecting = 0;
    INIT_CONNECT_TIMER(con, BT_TO_CONNECT_TOTAL);
    UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
  }
  con->flags = flags;

  return BT_ERR_NONE;

}

int BTHandlerGetFlags(cBTConnection con, int *flags)
{
  if (!con || !flags || con->fake)
    return BT_ERR_ARGUMENT;

  *flags = con->flags;

  DEBUG_LOG2("flags=0x%X, state=%s", *flags, STATE2STR(con->state));
  return BT_ERR_NONE;
}

int BTHandlerIsProfileActive(cBTConnection con, int profile, int *res)
{
  if (!con || !profile || con->fake)
    return BT_ERR_ARGUMENT;

  if (con->profile == profile && (con->state == BT_STATE_CONNECTED || con->state == BT_STATE_APP_READY))
    *res = BT_TRUE;
  else
    *res = BT_FALSE;

  return BT_ERR_NONE;
}

int BTHandlerNeedPair(cBTConnection con, int *res)
{
  if ((con->state == BT_STATE_START_LISTENING || con->state == BT_STATE_LISTENING ||
       con ->state == BT_STATE_START_CONNECTING || con->state == BT_STATE_CONNECTING ||
       con->state == BT_STATE_APP_START) && !con->fake)
    *res = BT_TRUE;
  else
    *res = BT_FALSE;

  return BT_ERR_NONE;
}

static void handle_connection_lost(BTConnection con)
{
  if (con->profile == BT_PROFILE_HF)
    DEBUG_LOG("handle_connection_lost : BT_PROFILE_HF disconnected"); 
   
  if (con->profile == BT_PROFILE_HS)
    DEBUG_LOG("handle_connection_lost : BT_PROFILE_HS disconnected");
   
  if (con->profile == BT_PROFILE_AVRCP)
    DEBUG_LOG("handle_connection_lost : BT_PROFILE_AVRCP disconnected");

  BTSocketClose(con->sock);
  
  // headset and avrcp profile can go to listen mode.
  if ((con->profile == BT_PROFILE_HS) || (con->profile == BT_PROFILE_AVRCP))  
  {
    if (con->flags & BT_FLAG_LISTEN) {
      DEBUG_LOG1("BT_FLAG_LISTEN flag is set => BT_STATE_START_LISTENING [timeout=%d]",BT_TO_LISTEN_TOTAL);
      INIT_LISTEN_TIMER(con, BT_TO_LISTEN_TOTAL);
      UPDATE_STATE(con->state, BT_STATE_START_LISTENING);
    } else {
      DEBUG_LOG("BT_FLAG_LISTEN flag not set => BT_STATE_INITIALIZED");
      UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
    }
  }
  else  // other profiles reconnect
  {
    con->reconnecting = 1;
    if (con->flags & BT_FLAG_DONOTRECONNECT) {
      DEBUG_LOG1("BT_FLAG_DONOTRECONNECT enabled, => BT_STATE_START_CONNECTING [timeout=%d]",BT_TO_CONNECT_TOTAL);
      INIT_CONNECT_TIMER(con, BT_TO_CONNECT_TOTAL);
    } else {
      DEBUG_LOG1("BT_FLAG_DONOTRECONNECT disabled, => BT_STATE_START_CONNECTING [timeout=%d]",BT_TO_RECONNECT_TOTAL);
      INIT_CONNECT_TIMER(con, BT_TO_RECONNECT_TOTAL);
    }
    UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
  }
}

static void handle_connect_failure(BTConnection con)
{
  if (BTTimerExpired(&con->connecttotaltr)) {
    if (con->flags & BT_FLAG_LISTEN) {
      DEBUG_LOG("number of retries exceeded, listening");
      INIT_LISTEN_TIMER(con, BT_TO_LISTEN_TOTAL);
      UPDATE_STATE(con->state, BT_STATE_START_LISTENING);
    } else {
      DEBUG_LOG("number of retries exceeded, initialized");
      UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
    }
  } else {
#ifndef NASHVILLE_BLUETOOTH
    int modelid = BTGetFileInt(BT_FILENAME_MODELID);

    if (!con->reconnecting &&  
        modelid != GOTYPE_VALENCIA500 && 
        modelid != GOTYPE_VALENCIA700 && 
        modelid != GOTYPE_MURCIA500 && 
        modelid != GOTYPE_MURCIA700 && 
        modelid != GOTYPE_MILAN && 
        modelid != GOTYPE_CAGLIARI)
    {
        DEBUG_LOG("reconnecting using short timeout");
        BTSetTimer(&con->connectretrytr, BT_TO_CONNECT_RETRY_1);
    } 
    else 
    {
        DEBUG_LOG("reconnecting using normal timeout");
        BTSetTimer(&con->connectretrytr, BT_TO_CONNECT_RETRY);
    }
#else
    // For NASHVILLE_BLUETOOTH use normal timeout.
    DEBUG_LOG("reconnecting using normal timeout");
    BTSetTimer(&con->connectretrytr, BT_TO_CONNECT_RETRY);
#endif

    if (con->state == BT_STATE_CONNECTING){
#ifdef FORCE_CH_SEARCH
      con->sock->chan = -1;
      DEBUG_LOG1(" connect error, set channel to %d - force search",con->sock->chan);
#endif
      UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
    }
  }
}

inline void empty_send_buffer(BTConnection con)
{
  if (con->sendbuf) {
    free(con->sendbuf);
    con->sendbuf = NULL;
    con->sendbuflen = 0;
  }
}

int hcitool_cc(bdaddr_t* bdaddr)
{
  uint16_t handle;
  uint8_t role;
  unsigned int ptype;
  int dev_id, dd, ret;
  struct hci_conn_info_req *cr;

  ret = 0;

  role = 0x01;
  ptype = HCI_DM1 | HCI_DM3 | HCI_DM5 | HCI_DH1 | HCI_DH3 | HCI_DH5;
  
  cr = malloc(sizeof(*cr) + sizeof(struct hci_conn_info));
  if (cr != NULL)  
  {
    dev_id = hci_get_route(bdaddr);
    if (dev_id >= 0)
    {
      dd = hci_open_dev(dev_id);
      if (dd >= 0)
      {
        //release all resources before we try to create a new baseband connection (prevents generic "Input/output error")
        bacpy(&cr->bdaddr, bdaddr);
        cr->type = ACL_LINK;
        if (ioctl(dd, HCIGETCONNINFO, (unsigned long) cr) != -1)
        {
          hci_disconnect(dd, htobs(cr->conn_info->handle), HCI_OE_USER_ENDED_CONNECTION, 10000);
        }

        //ok, lets create a new baseband connection
        DEBUG_LOG("Create baseband connection to remote device");
        if (hci_create_connection(dd, bdaddr, htobs(ptype), htobs(0x0000), role, &handle, 25000) >= 0)
        {
          DEBUG_LOG("Baseband connection established");
          sleep(1); //sleep one sec, to be 100% sure, that the baseband connection is really established
          ret = 1;
        }
        else
        {
          DEBUG_LOG1("Can't create connection: %s", strerror(errno));
        }
        hci_close_dev(dd);
      }
      else
      {
        DEBUG_LOG1("HCI device open failed: %s", strerror(errno));
      }
    }
    else
    {
      DEBUG_LOG1("Device is not available: %s", strerror(errno));
    }
    free(cr);
  }
  return ret;
}

int BTHandlerRun(BTConnection con, int disallow_connect)
{
  int i, res, res1, res2, res3, sent, pid;
  char* argv[10];
  char ttyname[BTS_TTY_NAME_LEN];
  char bdaddrstr[BT_BDADDR_LEN];
  struct stat s;
  pid_t ChildWait = -1, PPPChildWait = -1, HCIChildWait = -1;

  if (!con)
    return BT_ERR_ARGUMENT;
  switch (con->state) {
  case BT_STATE_UNINITIALIZED:
    if (con->fake)
      break;
    /* clean-up send buffer */
    empty_send_buffer(con);
    break;
  case BT_STATE_RETRIEVE_CHANNEL:

    con->chan_retrive_pid = fork();
    if (!con->chan_retrive_pid) {
      ba2str(&con->sock->bdaddr, bdaddrstr);
      res1 = BTHandlerScanProfileCH(bdaddrstr, con->profile);
      _exit(res1);

    } else if (con->chan_retrive_pid > 0) {
      DEBUG_LOG1("Parent: chiled started with pid=%d", con->chan_retrive_pid);
      UPDATE_STATE(con->state, BT_STATE_RETRIEVING_CHANNEL);
    } else {
      DEBUG_LOG("failed to start SDP query");
      UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
    }
    break;
  case BT_STATE_RETRIEVING_CHANNEL:

    if (con->chan_retrive_pid <= 0 ) {
      DEBUG_LOG("WRONG STATE!!! PLEASE CHECK THE STATE MACHINE!!!");
      UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
      break;
    }

    ChildWait = waitpid(con->chan_retrive_pid, &res1, WNOHANG);

    if (ChildWait > 0) {
      if (WIFEXITED(res1)){
        // Check receive channel value and verify that its not error
        if (WEXITSTATUS(res1)>0 && WEXITSTATUS(res1)<MAX_BT_CHANNEL_NUM) {
          con->sock->chan = WEXITSTATUS(res1);
          DEBUG_LOG1("set channel number [%d]", con->sock->chan);
          UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
          break;
        }
      }
      DEBUG_LOG1("sdp timeout expired, ChildWait [%d] ",ChildWait);
      if (con->sdpnrretries > 0) {
        con->sdpnrretries--;
        DEBUG_LOG1("Retrying SDP query, the number of attempts left is %d", con->sdpnrretries);
        UPDATE_STATE(con->state, BT_STATE_RETRIEVE_CHANNEL);
      } else {
        DEBUG_LOG("exceeded the number of retries for SDP query!");
        UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
      }
    }
    break;

  case BT_STATE_INITIALIZED:
    /* clean-up send buffer */
    empty_send_buffer(con);
    /* clean-up after pppd finished, sometimes it takes ages to bring it down */
    if (con->profile == BT_PROFILE_DUN)
    {
      if (con->ppp.pid > 0)
      {
        if (waitpid(con->ppp.pid, &res, WNOHANG))
        {
          con->ppp.pid = -1;
          BTSocketReleaseTTY(con->sock); 
          DEBUG_LOG("BT_STATE_INITIALIZED : PPPD finished, cleanup rfcomm"); 
        }
      }
    }
    break;
  case BT_STATE_START_LISTENING:
    /* clean-up send buffer */
    empty_send_buffer(con);
    /* put socket into the listening mode */
    if (!BTTimerExpired(&con->listenretrytr) && !BTTimerExpired(&con->listentotaltr))
      break;
    res = BTSocketListen(con->sock, (con->profile == BT_PROFILE_HF) ? BT_CHANNEL_HF : ( (con->profile == BT_PROFILE_HS) ? BT_CHANNEL_HS : con->sock->chan ) , con->profile);
    if (res != BT_ERR_NONE) {
      if (BTTimerExpired(&con->listentotaltr)) {
        DEBUG_LOG1("cannot start listening, error=%s", ERROR2STR(res));
        UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
      } else {
        BTSetTimer(&con->listenretrytr, BT_TO_LISTEN_RETRY);
      }
    } else {
      UPDATE_STATE(con->state, BT_STATE_LISTENING);
    }
    break;
  case BT_STATE_LISTENING:
    /* try to accept a new connection */
    res = BTSocketAccept(con->sock, BT_TO_NONE, con->profile);
    switch (res) {
    case BT_ERR_TIMEOUT:
      break;
    case BT_ERR_NONE:
      UPDATE_STATE(con->state, BT_STATE_CONNECTED);
      break;
    default:
      DEBUG_LOG1("cannot accept, error=%s", ERROR2STR(res));
      BTSocketStopListen(con->sock);
      BTSetTimer(&con->listenretrytr, BT_TO_LISTEN_RETRY);
      UPDATE_STATE(con->state, BT_STATE_START_LISTENING);
      break;
    }
    break;
  case BT_STATE_START_CONNECTING:
    if (disallow_connect)
      break;
    /* Connection management */
    if (con->fake) {
      UPDATE_STATE(con->state, BT_STATE_CONNECTING);
      INIT_CONNECT_TIMER(con, BT_TO_CONNECTIONREQUEST);
      break;
    }
    if (con->sock->chan < 0 && (con->profile != BT_PROFILE_AVRCP)) {
      con->sdpnrretries = 3;
      UPDATE_STATE(con->state, BT_STATE_RETRIEVE_CHANNEL);
      break;
    }
    /* put the socket into connecting mode */
    if (!BTTimerExpired(&con->connectretrytr) && !BTTimerExpired(&con->connecttotaltr))
      break;
    /* wait till ACL link is down (the hack needed by some phones) */
    if (!(con->flags & BT_FLAG_MULTICHANNEL)) {
      if (BTSocketGetAddress(con->sock, bdaddrstr) == BT_ERR_NONE) {
        if (BTIsAclLinkActive(bdaddrstr)) {
          DEBUG_LOG("ACL link is up, waiting...");
          BTSetTimer(&con->aclremovetr, BT_TO_REMOVEACL);
          UPDATE_STATE(con->state, BT_STATE_REMOVE_CONNECTION);
          break;
        }
      }
    }
    res = BTSocketConnect(con->sock, BT_TO_NONE, con->profile);
    switch (res) {
    case BT_ERR_NONE:
      UPDATE_STATE(con->state, BT_STATE_CONNECTED);
      break;
    case BT_ERR_TIMEOUT:
      UPDATE_STATE(con->state, BT_STATE_CONNECTING);
      break;
    default:
      DEBUG_LOG1("cannot start connecting, error=%s close the socket", ERROR2STR(res));
      handle_connect_failure(con);
      break;
    }
    break;
  case BT_STATE_REMOVE_CONNECTION:
    /* wait till ACL link is down (the hack needed by some phones) */
    if (BTSocketGetAddress(con->sock, bdaddrstr) == BT_ERR_NONE) {
      if (BTIsAclLinkActive(bdaddrstr) == BT_TRUE) {
        if (!BTTimerExpired(&con->aclremovetr))
          break;
        DEBUG_LOG("destroying ACL link.");
        BTDestroyAclLink(bdaddrstr);
      } else {
        DEBUG_LOG("ACL link is down.");
      }
    }
    BTSetTimer(&con->connectretrytr, 0);
    UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
    break;
  case BT_STATE_CONNECTING:
    /* Connection management */
    if (con->fake) {
      if (BTTimerExpired(&con->connecttotaltr)) {
        UPDATE_STATE(con->state, BT_STATE_UNINITIALIZED);
      }
      break;
    }
    /* check if connection is ready, no global timeout here - user may need to enter a PIN code for example */
    res = BTSocketIsConnected(con->sock, BT_TO_NONE);
    switch (res) {
    case BT_ERR_TIMEOUT:
      break;
    case BT_ERR_NONE:
      BTSetTimer(&con->pairattempttr, 0);
      BTSetTimer(&con->pairtotaltr, BT_TO_CHECK_PAIRING_TOTAL);
      UPDATE_STATE(con->state, BT_STATE_CHECK_PAIRING);
      break;
    default:
      DEBUG_LOG1("connecting failed, error=%s", ERROR2STR(res));
      BTSocketClose(con->sock);
      handle_connect_failure(con);
      break;
    }
    break;
  case BT_STATE_CHECK_PAIRING:
    if (!BTTimerExpired(&con->pairattempttr))
      break;
    ba2str(&con->sock->bdaddr, bdaddrstr);
    if (!BTIsPaired(bdaddrstr) && !BTTimerExpired(&con->pairtotaltr)) {
      BTSetTimer(&con->pairattempttr, BT_TO_CHECK_PAIRING_ATTEMPT);
      break;
    }
    UPDATE_STATE(con->state, BT_STATE_CONNECTED);
    break;
  case BT_STATE_CONNECTED:
    if (!con->sendbuf)
      break;
    res = BTSocketSend(con->sock, con->sendbuf, con->sendbuflen, &sent, BT_TO_NONE);
    if (res == BT_ERR_TIMEOUT)
      break;
    if (res != BT_ERR_NONE) {
      DEBUG_LOG("sending buffer: connection is lost");
      handle_connection_lost(con);
      break;
    }
    if (sent < con->sendbuflen) {
      DEBUG_LOG("send buffer partially sent");
      memcpy(con->sendbuf, con->sendbuf + sent, con->sendbuflen - sent);
      con->sendbuflen -= sent;
    } else {
      //DEBUG_LOG("send buffer purged");
      empty_send_buffer(con);
    }
    break;
  case BT_STATE_APP_START:
    if (disallow_connect)
      break;
    switch(con->ppp.state){

    case BT_PPP_RETRIEVE_CHANNEL:
      con->chan_retrive_pid_ppp = fork();
      if (!con->chan_retrive_pid_ppp) {
        ba2str(&con->sock->bdaddr, bdaddrstr);
        res2 = BTHandlerScanProfileCH(bdaddrstr, con->profile);
        DEBUG_LOG1("PPP Child retrieved channel [%d]",res2);
        _exit(res2);

      } else if (con->chan_retrive_pid_ppp > 0) {
        DEBUG_LOG1("Parent: PPP search channel chiled started with pid=%d", con->chan_retrive_pid_ppp);
        UPDATE_STATE(con->ppp.state, BT_PPP_RETRIEVING_CHANNEL);
      } else {
        DEBUG_LOG("failed to start PPP SDP query");
        UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
      }
      break;

    case BT_PPP_RETRIEVING_CHANNEL:
      if (con->chan_retrive_pid_ppp <= 0 ) {
        DEBUG_LOG("WRONG PPP STATE!!! PLEASE CHECK THE STATE MACHINE!!!");
        UPDATE_STATE(con->ppp.state, BT_PPP_STATE_INIT);
        break;
      }

      PPPChildWait = waitpid(con->chan_retrive_pid_ppp, &res2, WNOHANG);

      if (PPPChildWait > 0) {
        if (WIFEXITED(res2)){
          // Check receive channel value and verify that its not error
          if (WEXITSTATUS(res2)>0 && WEXITSTATUS(res2)<MAX_BT_CHANNEL_NUM) {
            con->sock->chan = WEXITSTATUS(res2);
            DEBUG_LOG1("set PPP channel number [%d] ", con->sock->chan);
            UPDATE_STATE(con->ppp.state, BT_PPP_STATE_CREATETTY);
            break;
          }
        }
        DEBUG_LOG1("sdp timeout expired, PPPChildWait [%d] ",ChildWait);

        if (con->sdpnrretries > 0) {
          con->sdpnrretries--;
          DEBUG_LOG1("Retrying SDP quesry for PPP , the number of attempts left is %d", con->sdpnrretries);
          UPDATE_STATE(con->ppp.state, BT_PPP_RETRIEVE_CHANNEL);
        } else {
          DEBUG_LOG("The number of retries for SDP query for PPP channel search exceeded!");
          UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
        }
      }
      break;
    case BT_PPP_STATE_HCITOOLCHECK:
      con->hcitool_pid = fork();
      if (!con->hcitool_pid)
      {
        if (hcitool_cc(&con->sock->bdaddr))
        {
          _exit(0);
        }
        else
        {
          sleep(1); //in case of an error, we sleep a bit
          _exit(1);
        }
      }
      else if (con->hcitool_pid > 0)
      {
        // code only executed by parent process
        DEBUG_LOG1("Parent: child started with pid=%d", con->hcitool_pid);
        //BTSetTimer(&con->hcitooltr, BT_TO_HCITOOLREADY);
        UPDATE_STATE(con->ppp.state, BT_PPP_STATE_HCITOOLCHECKREADY);
      }
      else
      {
        DEBUG_LOG("fork() failed");
        BTSocketReleaseTTY(con->sock);
        // Is this correct ?
        UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
      }
      break;

    case BT_PPP_STATE_HCITOOLCHECKREADY:
      if (con->hcitool_pid <= 0 )
      {
        DEBUG_LOG("WRONG PPP STATE!!! PLEASE CHECK THE STATE MACHINE!!!");
        BTSocketReleaseTTY(con->sock);
        UPDATE_STATE(con->ppp.state, BT_PPP_STATE_INIT);
        break;
      }

      HCIChildWait = waitpid(con->hcitool_pid, &res3, WNOHANG);
      //DEBUG_LOG2("returned %d, exit code %d", HCIChildWait, WEXITSTATUS(res3));
      if (HCIChildWait > 0 )
      {
        if (WIFEXITED(res3) && WEXITSTATUS(res3) == 0)
        {
          DEBUG_LOG1("Terminated normally, exit status =  %d", WEXITSTATUS(res3));
          // following three lines are copied from state BT_PPP_STATE_CHECKTTY just before fall through to BT_PPP_STATE_START
          UPDATE_STATE(con->ppp.state, BT_PPP_STATE_START);
          BTSetTimer(&con->ppp.startretrytr, 0);
          BTSetTimer(&con->ppp.starttotaltr, BT_TO_PPPSTARTTOTAL);
          break;
        }

        DEBUG_LOG1("wait hcitool_cc wait ended, ChildWait result [%d]",ChildWait);

        if (con->hcitoolnrretries > 0)
        {
          con->hcitoolnrretries--;
          DEBUG_LOG1("Retrying hcitool , the number of attempts left is %d", con->hcitoolnrretries);
          UPDATE_STATE(con->ppp.state, BT_PPP_STATE_HCITOOLCHECK);
        }
        else
        {
          DEBUG_LOG("The number of retries for hcitool exceeded!");
          BTSocketReleaseTTY(con->sock);
          //Is this correct?
          UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
        }
        break;

      }
      break;

    case BT_PPP_STATE_INIT:
      pid = BTGetFileInt(BT_PPP_PIDFILE);
      if (pid > 0) {
        DEBUG_LOG1("removing previous instance of %s", BT_PPP_NAME);
        kill(pid, SIGTERM);
        BTSetTimer(&con->ppp.removetr, BT_TO_APPREMOVE);
        UPDATE_STATE(con->ppp.state, BT_PPP_STATE_REMOVE);
        break;
      }
	  if (!(con->flags & BT_FLAG_MULTICHANNEL)&&!(con->flags & BT_FLAG_BREMEN_DEVICE)) {
        if (BTSocketGetAddress(con->sock, bdaddrstr) == BT_ERR_NONE) {
          if (BTIsAclLinkActive(bdaddrstr)) {
            DEBUG_LOG("ACL link is up, waiting.");
            BTSetTimer(&con->aclremovetr, BT_TO_REMOVEACL);
            UPDATE_STATE(con->state, BT_STATE_APP_REMOVE_CONNECTION);
            break;
          }
        }
      }
      UPDATE_STATE(con->ppp.state, BT_PPP_STATE_CREATETTY);
      //fall trough
    case BT_PPP_STATE_CREATETTY:
      if (con->sock->chan < 0) {
        con->sdpnrretries = 3;
        UPDATE_STATE(con->ppp.state, BT_PPP_RETRIEVE_CHANNEL);
        break;
      }
      BTSocketReleaseTTY(con->sock); //PLEASE CHECK: sometimes, /dev/rfcomm0 is already bound to an bluetooth address. 
      if (BTSocketCreateTTY(con->sock, BTS_TTY_DEFAULT_ID) != BT_ERR_NONE) {
        UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
        break;
      }
      BTSetTimer(&con->ppp.ttytr, BT_TO_TTYCREATE);
      UPDATE_STATE(con->ppp.state, BT_PPP_STATE_CHECKTTY);
      //fall though
    case BT_PPP_STATE_CHECKTTY:
      if (BTSocketIsTTYReady(con->sock, &res) != BT_ERR_NONE) {
        BTSocketReleaseTTY(con->sock);
        UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
        break;
      }
      if (res == BT_FALSE && !BTTimerExpired(&con->ppp.ttytr))
        break;
      if (res == BT_FALSE) {
        BTSocketReleaseTTY(con->sock);
        UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
        break;
      }
      DEBUG_LOG1("bdaddrstr is %s\n", bdaddrstr);

      //  workaround for bremen device
      if (con->flags & BT_FLAG_BREMEN_DEVICE)
      {
        con->hcitoolnrretries = 1;
        UPDATE_STATE(con->ppp.state, BT_PPP_STATE_HCITOOLCHECK);
        break;
      }

      UPDATE_STATE(con->ppp.state, BT_PPP_STATE_START);
      BTSetTimer(&con->ppp.startretrytr, 0);
      BTSetTimer(&con->ppp.starttotaltr, BT_TO_PPPSTARTTOTAL);
      //fall through
    case BT_PPP_STATE_START:
      /* put the socket into connecting mode */
      if (!BTTimerExpired(&con->ppp.startretrytr))
        break;
      i = 0;
      argv[i++] = BT_PPP_NAME;
      if (BTSocketGetTTYName(con->sock, ttyname) != BT_ERR_NONE) {
        DEBUG_LOG("failed to get TTY name");
        BTSocketReleaseTTY(con->sock);
        UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
        break;
      }
      //PLEASECHECK (Victor): work around, removing the PPP lock file
      unlink("/var/lock/LCK..rfcomm0");
      argv[i++] = ttyname;
      argv[i++] = BT_PPP_BAUDRATE;
#ifdef ENABLE_DEBUG_LOG
      argv[i++] = "debug";
#endif
      if (strlen(con->ppp.param)) {
        argv[i++] = "call";
        argv[i++] = con->ppp.param;
      }
      argv[i++] = "linkname";
      argv[i++] = BT_PPP_LINKNAME;
      argv[i++] = NULL;
      con->ppp.pid = vfork();
      if (!con->ppp.pid) {
        if (execvp(BT_PPP_NAME, argv) < 0) {
          DEBUG_LOG2("%s failed to start: %s", BT_PPP_NAME, strerror(errno));
          _exit(1);
        }
      } else if (con->ppp.pid > 0) {
        DEBUG_LOG2("%s started with pid=%d", BT_PPP_NAME, con->ppp.pid);
        BTSetTimer(&con->ppp.readytr, BT_TO_PPPREADY);
        UPDATE_STATE(con->ppp.state, BT_PPP_STATE_CHECKREADY);
      } else {
        DEBUG_LOG1("failed to start %s", BT_PPP_NAME);
        BTSocketReleaseTTY(con->sock);
        UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
      }
      break;
    case BT_PPP_STATE_CHECKREADY:
      if (BTTimerExpired(&con->ppp.readytr) || waitpid(con->ppp.pid, &res, WNOHANG)) {
        if (!BTTimerExpired(&con->ppp.readytr)) {
          if (WIFEXITED(res)) {
            DEBUG_LOG1("terminated normally, exit status =  %d", WEXITSTATUS(res));
            if (WEXITSTATUS(res) == 7 && !BTTimerExpired(&con->ppp.starttotaltr)) {
              DEBUG_LOG("retrying ppp startup: busy device");
              BTSetTimer(&con->ppp.startretrytr, BT_TO_PPPSTARTRETRY);
              UPDATE_STATE(con->ppp.state, BT_PPP_STATE_START);
              break;
            }
          }
          DEBUG_LOG1("%s failed after start", BT_PPP_NAME);
        } else {
          DEBUG_LOG1("%s timeout expired", BT_PPP_NAME);
        }
        BTSocketReleaseTTY(con->sock);
        pid = BTGetFileInt(BT_PPP_PIDFILE);
        if (pid > 0) {
          DEBUG_LOG1("terminating %d", pid);
          kill(pid, SIGTERM);
        }
        BTSetTimer(&con->ppp.removetr, BT_TO_APPREMOVE);
        UPDATE_STATE(con->state, BT_STATE_APP_CLOSING);
      } else if (stat(BT_PPP_UPFILE, &s) >= 0 ) {
#ifndef NASHVILLE_BLUETOOTH
        BTSocketReleaseTTY(con->sock); // This causes DUN problems for Nashville
#endif
        UPDATE_STATE(con->state, BT_STATE_APP_READY);
      }
      break;
    case BT_PPP_STATE_REMOVE:
      pid = BTGetFileInt(BT_PPP_PIDFILE);
      if (pid > 0) {
        if (!waitpid(pid, &res, WNOHANG) && !BTTimerExpired(&con->ppp.removetr))
          break;
        pid = BTGetFileInt(BT_PPP_PIDFILE);
        if (pid > 0) {
          DEBUG_LOG1("sending KILL to %s", BT_PPP_NAME);
          kill(pid, SIGKILL);
          waitpid(pid, &res, WNOHANG);
        }
      }
      UPDATE_STATE(con->ppp.state, BT_PPP_STATE_INIT);
      break;
    default:
      break;
    }
    break;
  case BT_STATE_APP_REMOVE_CONNECTION:
    if (BTSocketGetAddress(con->sock, bdaddrstr) == BT_ERR_NONE) {
      if (BTIsAclLinkActive(bdaddrstr) == BT_TRUE) {
        if (!BTTimerExpired(&con->aclremovetr))
          break;
        DEBUG_LOG("destroying ACL link.");
        BTDestroyAclLink(bdaddrstr);
      } else {
        DEBUG_LOG("ACL link is down.");
      }
    }
    UPDATE_STATE(con->state, BT_STATE_APP_START);
    break;
  case BT_STATE_APP_READY:
    if (waitpid(con->ppp.pid, &res, WNOHANG))
    {
      UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
    }
    break;
  case BT_STATE_APP_CLOSING:
    if (BTTimerExpired(&con->ppp.removetr) || waitpid(con->ppp.pid, &res, WNOHANG)) 
    {
      if (!BTTimerExpired(&con->ppp.removetr))
      {
        con->ppp.pid = -1;
        BTSocketReleaseTTY(con->sock);
        DEBUG_LOG("BT_STATE_APP_CLOSING : PPPD finished, cleanup rfcomm"); 
      }
      UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
    }
    break;
  case BT_STATE_CLOSING:
    if (con->chan_retrive_pid > 0)
    {
      DEBUG_LOG1("send kill to pid [%d] ",con->chan_retrive_pid)
          kill(con->chan_retrive_pid, SIGTERM);
      ChildWait = waitpid(con->chan_retrive_pid, &res1, WNOHANG);
      if (!PPPChildWait)
         break;
    }
	if (con->hcitool_pid > 0)
    {
      DEBUG_LOG1("send kill to hcitool pid - [%d] ",con->hcitool_pid)
      kill(con->hcitool_pid, SIGTERM);
      HCIChildWait = waitpid(con->hcitool_pid, &res3, WNOHANG);

      if (!ChildWait)
        break;
    }
    if (con->chan_retrive_pid_ppp > 0)
    {
      DEBUG_LOG1("send kill to pid - (ppp) [%d] ",con->chan_retrive_pid_ppp)
          kill(con->chan_retrive_pid_ppp, SIGTERM);
      PPPChildWait = waitpid(con->chan_retrive_pid_ppp, &res1, WNOHANG);

      if (!ChildWait)
        break;
    }

    switch(con->ppp.state){
    case BT_PPP_STATE_CLOSING:
      if (BTTimerExpired(&con->ppp.removetr) || waitpid(con->ppp.pid, &res, WNOHANG)) 
      {
        if (!BTTimerExpired(&con->ppp.removetr))
        {
          con->ppp.pid = -1;
          BTSocketReleaseTTY(con->sock);
          DEBUG_LOG("BT_STATE_CLOSING BT_PPP_STATE_CLOSING : PPPD finished, cleanup rfcomm"); 
        }
        UPDATE_STATE(con->state, BT_STATE_UNINITIALIZED);
      }
      break;
    case BT_PPP_STATE_UNINITIALIZED:
    default:
      UPDATE_STATE(con->state, BT_STATE_UNINITIALIZED);
      break;
    }
    break;
  default:
    break;

  }
  return BT_ERR_NONE;
}

int BTHandlerGetBTName(char* name, int size)
{
  int result;

  if (BTGetName(name, size))
    result = BT_ERR_FAILED;
  else
    result = BT_ERR_NONE;

  return result;
}

int BTHandlerSetBTName(char* name)
{
  int result;

  if (BTSetName(name))
    result = BT_ERR_FAILED;
  else
    result = BT_ERR_NONE;

  return result;
}

int BTHandlerExitSniffMode(char* bdaddr)
{
  int result;
  DEBUG_LOG1("BTHandlerExitSniffMode[%s] ",bdaddr);
  if (BTExitSniffMode(bdaddr))
  {
    result = BT_ERR_FAILED;
    DEBUG_LOG("bthandler:BTHandlerExitSniffMode | failed");
  }
  else
  {
    result = BT_ERR_NONE;
    DEBUG_LOG("bthandler:BTHandlerExitSniffMode | succeeded");
  }
    
  return result;
}

// SDP query - get channel number form adress by profile
int BTHandlerScanProfileCH(const char *bAdress, int SearchProfile)
{
  DEBUG_LOG2("search channel for address [%s], profile [%d]",bAdress, SearchProfile);
  sdp_session_t *sess = 0;
  sdp_list_t *attrid, *SearchList, *seq, *proto;
  sdp_list_t *next = 0;
  uint32_t range = 0x0000ffff;
  int channel = -1;
  bdaddr_t SktAdress;
  uuid_t ContextGroup;

  //set the real values for profile
  switch (SearchProfile){
  case BT_PROFILE_HF:
    DEBUG_LOG("HANDSFREE_SVCLASS_ID search for channel");
    SearchProfile = 0x111f;
    break;
  case BT_PROFILE_HS:
    DEBUG_LOG("HEADSET_SVCLASS_ID search for channel");
    SearchProfile = 0x1112;
    break;
  case BT_PROFILE_DUN:
    DEBUG_LOG("DIALUP_NET_SVCLASS_ID search for channel");
    SearchProfile = 0x1103;
    break;
  case BT_PROFILE_AVRCP:  // ACRCP does not require a channel nr as it uses L2CAP
    DEBUG_LOG("Error: BTHandlerScanProfileCH should not be called for AVRCP!");
    break;
  case BT_PROFILE_SP:
  default:
    DEBUG_LOG1("Profile (=%d) search for SERIAL PORT channel",SearchProfile);
    SearchProfile = 0x1101;
  }

  str2ba(bAdress,&SktAdress);
  // test if device alive and connect
  DEBUG_LOG1(" Run sdp_connect to %s", bAdress);
  sess = sdp_connect(BDADDR_ANY, &SktAdress, SDP_RETRY_IF_BUSY);
  if (!sess) {
    char addr[20];
    ba2str(&SktAdress,addr);
    DEBUG_LOG1("Failed to connect to SDP server on %s",addr );
    return -1;
  }
  sdp_uuid16_create(&ContextGroup, SearchProfile);
  attrid = sdp_list_append(0, &range);

  SearchList = sdp_list_append(0, &ContextGroup);

  // get a list of service records
  if (sdp_service_search_attr_req(sess, SearchList, SDP_ATTR_REQ_RANGE, attrid, &seq)) {
    DEBUG_LOG1("Service Search failed: %s", strerror(errno));
    sdp_close(sess);
    return -1;
  }
  sdp_list_free(attrid, 0);
  sdp_list_free(SearchList, 0);
  // go through each of the service records
  for (; seq; seq = next){
    sdp_record_t *rec = (sdp_record_t *) seq->data;
    //Get channel info
    if (sdp_get_access_protos(rec, &proto) == 0)
      channel = sdp_get_proto_port(proto, RFCOMM_UUID);

    next = seq->next;
    free(seq);
    sdp_record_free(rec);
  }

  sdp_list_free(proto, 0);
  sdp_list_free(next, 0);
  sdp_close(sess);

  if (channel > 0)
    return channel;

  DEBUG_LOG("Fail to find channel");
  return BT_ERR_FAILED;
}

int BTHandlerGetMtuSize(cBTConnection con, int * imtu, int * omtu)
{
  if (!con || con->fake)
    return BT_ERR_ARGUMENT;

  return BTSocketGetMtuSize(con->sock, imtu, omtu);
}

int BTHandlerCreateTTY(cBTConnection con, int tty_id)
{
  if (!con || con->fake || !con->sock)
    return BT_ERR_ARGUMENT;

  return BTSocketCreateTTY(con->sock, tty_id);
}

int BTHandlerReleaseTTY(cBTConnection con)
{
  if (!con || con->fake || !con->sock)
    return BT_ERR_ARGUMENT;

  return BTSocketReleaseTTY(con->sock);
}
